import getpass
import logging
import mastodon
from .base import Silo, upload_silo_media


logger = logging.getLogger(__name__)


class MastodonSilo(Silo):
    SILO_TYPE = 'mastodon'
    _CLIENT_CLASS = mastodon.Mastodon

    def __init__(self, ctx):
        super().__init__(ctx)

        self.base_url = self.getConfigItem('url')
        if not self.base_url:
            raise Exception("No Mastodon instance URL defined for: %s",
                            self.ctx.silo_name)

        self.client = None

    def authenticate(self, ctx):
        force = ctx.exec_ctx.args.force

        client_token = self.getCacheItem('clienttoken')
        if not client_token or force:
            logger.info("Authenticating client app with Mastodon for %s" %
                        self.ctx.silo_name)
            pair = self._CLIENT_CLASS.create_app(
                'SiloRider',
                scopes=['read', 'write'],
                api_base_url=self.base_url
            )
            client_token = '%s,%s' % pair
            self.setCacheItem('clienttoken', client_token)

        client_id, client_secret = client_token.split(',')

        access_token = self.getCacheItem('accesstoken')
        if not access_token or force:
            m = self._CLIENT_CLASS(
                client_id=client_id,
                client_secret=client_secret,
                api_base_url=self.base_url)

            if ctx.exec_ctx.args.console:
                logger.info("Authenticating user with Mastodon for %s" %
                            self.ctx.silo_name)
                logger.info("Only access tokens will be stored -- your "
                            "username and password will be forgotten in "
                            "a second.")

                username = input("Username: ")
                if not username:
                    raise Exception("You must enter a username.")

                password = getpass.getpass(prompt="Password: ")

                try:
                    access_token = m.log_in(
                        username, password,
                        scopes=['read', 'write'])
                except mastodon.MastodonIllegalArgumentError as err:
                    raise Exception("Incorrect credientials") from err
                except mastodon.MastodonAPIError as err:
                    raise Exception("Autentication error") from err

                username = password = None

            else:
                logger.info("Once you've authorized silorider to access"
                            "your Mastodon account, paste the authentication "
                            "code back here:")

                import webbrowser
                req_url = m.auth_request_url(scopes=['write'])
                webbrowser.open(req_url)

                access_token = input("Authentication code: ")

            self.setCacheItem('accesstoken', access_token)

    def onPostStart(self):
        self._ensureApp()

    def _ensureApp(self):
        if self.client is not None:
            return

        logger.debug("Creating Mastodon app.")
        client_token = self.getCacheItem('clienttoken')
        if not client_token:
            raise Exception("Mastodon silo '%s' isn't authenticated." %
                            self.name)

        client_id, client_secret = client_token.split(',')

        access_token = self.getCacheItem('accesstoken')
        if not access_token:
            raise Exception("Mastodon silo '%s' isn't authenticated." %
                            self.name)

        self.client = self._CLIENT_CLASS(
            client_id=client_id,
            client_secret=client_secret,
            access_token=access_token,
            api_base_url=self.base_url)

    def postEntry(self, entry, ctx):
        toottxt = self.formatEntry(entry, limit=500)
        if not toottxt:
            raise Exception("Can't find any content to use for the toot!")

        visibility = self.getConfigItem('toot_visibility', fallback='public')

        media_ids = upload_silo_media(entry, 'photo', self._media_callback)

        logger.debug("Posting toot: %s" % toottxt)
        self.client.status_post(toottxt, media_ids=media_ids,
                                visibility=visibility)

    def _media_callback(self, tmpfile, mt):
        with open(tmpfile, 'rb') as tmpfp:
            return self.client.media_post(tmpfp, mt)



import pytest


def test_one_article(cli, feedutil, tweetmock):
    feed = cli.createTempFeed(feedutil.makeFeed(
        """<h1 class="p-name">A new article</h1>
<div class="e-content">
<p>This is the text of the article.</p>
<p>It has 2 paragraphs.</p>
</div>
<a class="u-url" href="https://example.org/a-new-article">permalink</a>"""
    ))

    cli.appendSiloConfig('test', 'twitter', url='/blah')
    tweetmock.installTokens(cli, 'test')

    ctx, _ = cli.run('process', feed)
    assert ctx.cache.wasPosted('test', 'https://example.org/a-new-article')
    toot = ctx.silos[0].client.tweets[0]
    assert toot == ('A new article https://example.org/a-new-article', [])


def test_one_micropost(cli, feedutil, tweetmock):
    feed = cli.createTempFeed(feedutil.makeFeed(
        """<p class="p-name">This is a quick update.</p>
<a class="u-url" href="/01234.html">permalink</a>"""
    ))

    cli.appendSiloConfig('test', 'twitter', url='/blah')
    tweetmock.installTokens(cli, 'test')

    ctx, _ = cli.run('process', feed)
    assert ctx.cache.wasPosted('test', '/01234.html')
    toot = ctx.silos[0].client.tweets[0]
    assert toot == ("This is a quick update.", [])


def test_one_micropost_with_one_photo(cli, feedutil, tweetmock, monkeypatch):
    feed = cli.createTempFeed(feedutil.makeFeed(
        """<p class="p-name">This is a quick photo update.</p>
<div>
    <a class="u-photo" href="/fullimg.jpg"><img src="/thumbimg.jpg"/></a>
</div>
<a class="u-url" href="/01234.html">permalink</a>"""
    ))

    cli.appendSiloConfig('test', 'twitter', url='/blah')
    tweetmock.installTokens(cli, 'test')

    ctx, _ = cli.run('process', feed)

    assert ctx.cache.wasPosted('test', '/01234.html')
    toot = ctx.silos[0].client.tweets[0]
    assert toot == ("This is a quick photo update.", ['/fullimg.jpg'])


def test_one_micropost_with_two_photos(cli, feedutil, tweetmock, monkeypatch):
    feed = cli.createTempFeed(feedutil.makeFeed(
        """<p class="p-name">This is a photo update with 2 photos.</p>
<div>
    <a class="u-photo" href="/fullimg1.jpg"><img src="/thumbimg1.jpg"/></a>
    <a class="u-photo" href="/fullimg2.jpg"><img src="/thumbimg2.jpg"/></a>
</div>
<a class="u-url" href="/01234.html">permalink</a>"""
    ))

    cli.appendSiloConfig('test', 'twitter', url='/blah')
    tweetmock.installTokens(cli, 'test')

    ctx, _ = cli.run('process', feed)

    assert ctx.cache.wasPosted('test', '/01234.html')
    toot = ctx.silos[0].client.tweets[0]
    assert toot == ("This is a photo update with 2 photos.",
                    ['/fullimg1.jpg', '/fullimg2.jpg'])


@pytest.fixture(scope='session')
def tweetmock():
    from silorider.silos.twitter import TwitterSilo
    TwitterSilo._CLIENT_CLASS = TwitterMock
    return TwitterMockUtil()


class TwitterMock:
    def __init__(self, consumer_key, consumer_secret,
                 access_token_key, access_token_secret):
        assert consumer_key == 'TEST_CLIENT_KEY'
        assert consumer_secret == 'TEST_CLIENT_SECRET'
        assert access_token_key == 'TEST_ACCESS_KEY'
        assert access_token_secret == 'TEST_ACCESS_SECRET'

        self.tweets = []

    def PostUpdate(self, tweet, media=None):
        self.tweets.append((tweet, media))


class TwitterMockUtil:
    def installTokens(self, cli, silo_name):
        def do_install_tokens(ctx):
            ctx.cache.setCustomValue(
                '%s_clienttoken' % silo_name,
                'TEST_CLIENT_KEY,TEST_CLIENT_SECRET')
            ctx.cache.setCustomValue(
                '%s_accesstoken' % silo_name,
                'TEST_ACCESS_KEY,TEST_ACCESS_SECRET')

        cli.preExecHook(do_install_tokens)



import os.path
import urllib.parse
import logging


logger = logging.getLogger(__name__)


class Cache:
    def getCustomValue(self, name, valtype=str):
        raise NotImplementedError()

    def setCustomValue(self, name, val):
        raise NotImplementedError()

    def wasPosted(self, silo_name, entry_uri):
        raise NotImplementedError()

    def addPost(self, silo_name, entry_uri):
        raise NotImplementedError()


class NullCache(Cache):
    def __init__(self):
        self._vals = {}

    def getCustomValue(self, name, valtype=str):
        return self._vals.get(name)

    def setCustomValue(self, name, val):
        self._vals[name] = val

    def wasPosted(self, silo_name, entry_uri):
        return False

    def addPost(self, silo_name, entry_uri):
        pass


def load_cache(config, cfg_dir):
    if not config.has_section('cache'):
        logger.warning("No cache configured!")
        return NullCache()

    cache_uri = config.get('cache', 'uri', fallback=None)
    if not cache_uri:
        return NullCache()

    res = urllib.parse.urlparse(cache_uri)
    if res.scheme == 'sqlite':
        from .sqlite import SqliteCache
        dbpath = res.netloc + res.path
        if cfg_dir:
            dbpath = os.path.join(cfg_dir, dbpath)
        return SqliteCache(dbpath, config)
    elif res.scheme == 'memory':
        from .memory import MemoryCache
        return MemoryCache()

    raise Exception("Unknown cache URI: %s" % cache_uri)



import os.path
import logging
from .base import Cache


logger = logging.getLogger(__name__)


class SqliteCache(Cache):
    SCHEMA_VERSION = 1

    def __init__(self, dbpath, config):
        self.path = dbpath
        self.config = config

        import sqlite3
        logger.debug("Opening SQL DB: %s" % dbpath)
        self.conn = sqlite3.connect(dbpath,
                                    detect_types=sqlite3.PARSE_DECLTYPES)

        if (not os.path.exists(dbpath) or
                self._getSchemaVersion() != self.SCHEMA_VERSION):
            self._initDb()

    def _getSchemaVersion(self):
        import sqlite3
        try:
            return self.getCustomValue('schema_version', valtype=int)
        except sqlite3.Error:
            return None

    def _initDb(self):
        c = self.conn.cursor()
        c.execute('''DROP TABLE IF EXISTS info''')
        c.execute(
            '''CREATE TABLE info (
                name text PRIMARY KEY NOT NULL,
                str_val text,
                real_val real,
                int_val int
            )''')
        c.execute(
            '''INSERT INTO info (name, int_val)
                VALUES ('schema_version', ?)''',
            (self.SCHEMA_VERSION,))

        c.execute('''DROP TABLE IF EXISTS posted''')
        c.execute(
            '''CREATE TABLE posted (
                id integer PRIMARY KEY,
                silo text NOT NULL,
                uri text NOT NULL,
                posted_on timestamp
            )''')
        c.execute(
            '''CREATE INDEX index_silo ON posted(silo)''')
        c.execute(
            '''CREATE INDEX index_uri ON posted(uri)''')
        self.conn.commit()
        c.close()

    def getCustomValue(self, name, valtype=str):
        c = self.conn.cursor()
        if valtype is str:
            c.execute(
                '''SELECT str_val FROM info WHERE (name = ?)''', (name,))
        elif valtype is float:
            c.execute(
                '''SELECT real_val FROM info WHERE (name = ?)''', (name,))
        elif valtype in (int, bool):
            c.execute(
                '''SELECT int_val FROM info WHERE (name = ?)''', (name,))
        else:
            raise Exception("Unsupported value type: %s" % valtype)
        row = c.fetchone()
        if row is None:
            return None

        return valtype(row[0])

    def setCustomValue(self, name, val):
        c = self.conn.cursor()
        if isinstance(val, str):
            c.execute(
                '''INSERT OR REPLACE INTO info (name, str_val)
                    VALUES (?, ?)''',
                (name, val))
        elif isinstance(val, float):
            c.execute(
                '''INSERT OR REPLACE INTO info (name, real_val)
                    VALUES (?, ?)''',
                (name, str(val)))
        elif isinstance(val, (int, bool)):
            c.execute(
                '''INSERT OR REPLACE INTO info (name, int_val)
                    VALUES (?, ?)''',
                (name, str(int(val))))
        else:
            raise Exception("Unsupported value type: %s" % type(val))

        self.conn.commit()
        c.close()

    def wasPosted(self, silo_name, entry_uri):
        c = self.conn.cursor()
        c.execute(
            '''SELECT id, silo, uri
                FROM posted
                WHERE (silo = ? AND uri = ?)''',
            (silo_name, entry_uri))
        if c.fetchone():
            return True
        return False

    def addPost(self, silo_name, entry_uri):
        c = self.conn.cursor()
        c.execute(
            '''INSERT INTO posted (silo, uri)
                VALUES (?, ?)''',
            (silo_name, entry_uri))
        self.conn.commit()
        c.close()



import pytest
from .mockutil import mock_urllib


def test_one_article(cli, feedutil, mastmock):
    feed = cli.createTempFeed(feedutil.makeFeed(
        """<h1 class="p-name">A new article</h1>
<div class="e-content">
<p>This is the text of the article.</p>
<p>It has 2 paragraphs.</p>
</div>
<a class="u-url" href="https://example.org/a-new-article">permalink</a>"""
    ))

    cli.appendSiloConfig('test', 'mastodon', url='/blah')
    mastmock.installTokens(cli, 'test')

    ctx, _ = cli.run('process', feed)
    assert ctx.cache.wasPosted('test', 'https://example.org/a-new-article')
    toot = ctx.silos[0].client.toots[0]
    assert toot == ('A new article https://example.org/a-new-article',
                    None, 'public')


def test_one_micropost(cli, feedutil, mastmock):
    feed = cli.createTempFeed(feedutil.makeFeed(
        """<p class="p-name">This is a quick update.</p>
<a class="u-url" href="/01234.html">permalink</a>"""
    ))

    cli.appendSiloConfig('test', 'mastodon', url='/blah')
    mastmock.installTokens(cli, 'test')

    ctx, _ = cli.run('process', feed)
    assert ctx.cache.wasPosted('test', '/01234.html')
    toot = ctx.silos[0].client.toots[0]
    assert toot == ("This is a quick update.", None, 'public')


def test_one_micropost_with_one_photo(cli, feedutil, mastmock, monkeypatch):
    feed = cli.createTempFeed(feedutil.makeFeed(
        """<p class="p-name">This is a quick photo update.</p>
<div>
    <a class="u-photo" href="/fullimg.jpg"><img src="/thumbimg.jpg"/></a>
</div>
<a class="u-url" href="/01234.html">permalink</a>"""
    ))

    cli.appendSiloConfig('test', 'mastodon', url='/blah')
    mastmock.installTokens(cli, 'test')

    with monkeypatch.context() as m:
        import silorider.silos.mastodon
        mock_urllib(m)
        m.setattr(silorider.silos.mastodon.MastodonSilo, '_media_callback',
                  _patched_media_callback)
        ctx, _ = cli.run('process', feed)

    assert ctx.cache.wasPosted('test', '/01234.html')
    media = ctx.silos[0].client.media[0]
    assert media == ('/retrieved/fullimg.jpg', 'image/jpeg', 1)
    toot = ctx.silos[0].client.toots[0]
    assert toot == ("This is a quick photo update.", [1], 'public')


def test_one_micropost_with_two_photos(cli, feedutil, mastmock, monkeypatch):
    feed = cli.createTempFeed(feedutil.makeFeed(
        """<p class="p-name">This is a photo update with 2 photos.</p>
<div>
    <a class="u-photo" href="/fullimg1.jpg"><img src="/thumbimg1.jpg"/></a>
    <a class="u-photo" href="/fullimg2.jpg"><img src="/thumbimg2.jpg"/></a>
</div>
<a class="u-url" href="/01234.html">permalink</a>"""
    ))

    cli.appendSiloConfig('test', 'mastodon', url='/blah')
    mastmock.installTokens(cli, 'test')

    with monkeypatch.context() as m:
        import silorider.silos.mastodon
        mock_urllib(m)
        m.setattr(silorider.silos.mastodon.MastodonSilo, '_media_callback',
                  _patched_media_callback)
        ctx, _ = cli.run('process', feed)

    assert ctx.cache.wasPosted('test', '/01234.html')
    media = ctx.silos[0].client.media[0]
    assert media == ('/retrieved/fullimg1.jpg', 'image/jpeg', 1)
    media = ctx.silos[0].client.media[1]
    assert media == ('/retrieved/fullimg2.jpg', 'image/jpeg', 2)
    toot = ctx.silos[0].client.toots[0]
    assert toot == ("This is a photo update with 2 photos.", [1, 2], 'public')


def _patched_media_callback(self, tmpfile, mt):
    return self.client.media_post(tmpfile, mt)


@pytest.fixture(scope='session')
def mastmock():
    from silorider.silos.mastodon import MastodonSilo
    MastodonSilo._CLIENT_CLASS = MastodonMock
    return MastodonMockUtil()


class MastodonMock:
    @staticmethod
    def create_app(app_name, scopes, api_base_url):
        return ('TEST_CLIENT_ID', 'TEST_CLIENT_SECRET')

    def __init__(self, client_id, client_secret, access_token, api_base_url):
        self.toots = []
        self.media = []
        self.next_mid = 1

    def log_in(self, username, password, scopes):
        return 'TEST_ACCESS_TOKEN'

    def auth_request_url(self, scopes):
        return 'https://example.org/auth'

    def status_post(self, toot, media_ids=None, visibility=None):
        self.toots.append((toot, media_ids, visibility))

    def media_post(self, filename, mimetype):
        mid = self.next_mid
        self.next_mid += 1
        self.media.append((filename, mimetype, mid))
        return mid


class MastodonMockUtil:
    def installTokens(self, cli, silo_name):
        def do_install_tokens(ctx):
            ctx.cache.setCustomValue(
                '%s_clienttoken' % silo_name,
                'TEST_CLIENT_ID,TEST_CLIENT_SECRET')
            ctx.cache.setCustomValue(
                '%s_accesstoken' % silo_name,
                'TEST_ACCESS_TOKEN')

        cli.preExecHook(do_install_tokens)



import logging
from ..parse import parse_mf2


logger = logging.getLogger(__name__)


def get_named_silos(silos, names):
    if not names:
        return silos

    valid_names = set([s.name for s in silos])
    for n in names:
        if n not in valid_names:
            raise Exception("No such silo: %s" % n)

    res = []
    for s in silos:
        if s.name in names:
            res.append(s)
    return res


def populate_cache(url, ctx):
    import mf2util
    import dateutil.parser

    silos = get_named_silos(ctx.silos, ctx.args.silo)

    until_dt = None
    if ctx.args.until:
        until_dt = dateutil.parser.parse(ctx.args.until).date()
        logger.debug("Populating cache until: %s" % until_dt)

    mf_obj = parse_mf2(url)
    mf_dict = mf_obj.to_dict()
    for entry in mf_dict.get('items', []):
        entry_props = entry.get('properties')
        if not entry_props:
            logger.warning("Found entry without any properties.")
            continue

        entry_url = entry_props.get('url')
        if not entry_url:
            logger.warning("Found entry without any URL.")
            continue

        if isinstance(entry_url, list):
            entry_url = entry_url[0]

        if until_dt:
            entry_published = entry_props.get('published')
            if not entry_published:
                logger.warning("Entry '%s' has not published date." %
                               entry_url)
                continue

            if isinstance(entry_published, list):
                entry_published = entry_published[0]

            entry_published_dt = mf2util.parse_datetime(entry_published)
            if entry_published_dt and entry_published_dt.date() > until_dt:
                continue

        logger.debug("Adding entry to cache: %s" % entry_url)
        for silo in silos:
            ctx.cache.addPost(silo.name, entry_url)



import os.path
import logging
import datetime


logger = logging.getLogger(__name__)


def parse_url(url_or_path):
    mf_obj = parse_mf2(url_or_path)
    matcher = EntryMatcher(mf_obj)

    feed = Feed(url_or_path, matcher.mf_dict)

    entries = []
    for pair in matcher.entries:
        mf_entry, bs_el = pair
        try:
            entry = Entry(feed, mf_entry, bs_el)
            entry.interpret()
        except InvalidEntryException:
            logger.debug("Found invalid entry... skipping.")
            continue

        entries.append(entry)

    sorted_entries = sorted(
        entries,
        key=lambda e: e.get(
            'published', datetime.datetime.fromtimestamp(
                0,
                tz=datetime.timezone(datetime.timedelta(0)))),
        reverse=False)

    feed.entries = sorted_entries
    return feed


def parse_mf2(url_or_path):
    import mf2py
    logger.debug("Fetching %s..." % url_or_path)
    if os.path.exists(url_or_path):
        obj = open(url_or_path, 'r', encoding='utf8')
        params = {'doc': obj}
    else:
        params = {'url': url_or_path}
    return mf2py.Parser(html_parser='html5lib', **params)


class InvalidEntryException(Exception):
    pass


class Feed:
    def __init__(self, url, mf_dict):
        self.url = url
        self._mf_dict = mf_dict
        self.entries = []


class Entry:
    def __init__(self, owner_feed, mf_entry, bs_obj):
        self._owner_feed = owner_feed
        self._mf_entry = mf_entry
        self._bs_obj = bs_obj

        self._type = None
        self._props = None

    @property
    def entry_type(self):
        return self._type

    @property
    def html_element(self):
        return self._bs_obj

    @property
    def best_name(self):
        self.interpret()

        for pn in ['title', 'name', 'content-plain', 'content']:
            pv = self._props.get(pn)
            if pv:
                return pv
        return None

    def __getattr__(self, name):
        try:
            return self._doGet(name)
        except KeyError:
            raise AttributeError("Entry does not have property '%s'." % name)

    def get(self, name, default=None, *, force_list=False):
        try:
            return self._doGet(name, force_list=force_list)
        except KeyError:
            return default

    def _doGet(self, name, force_list=False):
        self.interpret()

        values = self._props[name]
        if not force_list and isinstance(values, list) and len(values) == 1:
            return values[0]
        return values

    def htmlFind(self, *args, **kwargs):
        if self._bs_obj is None:
            raise Exception("No HTML object is available for this entry.")

        return self._bs_obj.find(*args, **kwargs)

    def interpret(self):
        if self._type is not None or self._props is not None:
            return

        import mf2util

        self._type = mf2util.post_type_discovery(self._mf_entry)
        self._props = mf2util.interpret_entry(
            self._owner_feed._mf_dict, self._owner_feed.url,
            hentry=self._mf_entry)

        # Adds a `is_micropost` property.
        self._detect_micropost()

        # mf2util only detects the first photo for a "photo"-type post,
        # but there might be several so we need to fix that.
        #
        # mf2util also apparently doesn't always bring "category" info.
        self._fix_interpreted_props('photo', 'category')

    def _detect_micropost(self):
        is_micro = False
        name = self.get('name')
        content = self.get('content-plain')
        if content and not name:
            is_micro = True
        elif name and not content:
            is_micro = True
        elif name and content:
            shortest = min(len(name), len(content))
            is_micro = (name[:shortest] == content[:shortest])
        self._props['is_micropost'] = is_micro

    def _fix_interpreted_props(self, *names):
        for name in names:
            values = self._mf_entry['properties'].get(name, [])
            if isinstance(values, str):
                values = [values]
            self._props[name] = values


class EntryMatcher:
    """ A class that matches `mf2util` results along with the original
        BeautifulSoup document, so we have HTML objects on hand if needed.
    """
    def __init__(self, mf_obj):
        self.mf_dict = mf_obj.to_dict()
        self.entries = []

        els_by_type = {}
        next_el = {}
        bf = mf_obj.__doc__
        for e in self.mf_dict.get('items', []):
            types = e.get('type')
            if not types:
                continue

            entry_type = types[0]
            if entry_type not in els_by_type:
                ebt = list(bf.find_all(class_=entry_type))
                els_by_type[entry_type] = ebt
                next_el[entry_type] = 0

            els = els_by_type[entry_type]
            e_and_el = (e, els[next_el[entry_type]])
            self.entries.append(e_and_el)
            next_el[entry_type] += 1



import pytest
from silorider.format import format_entry


test_url = 'https://example.org/article'


class TestEntry:
    pass


def _make_test_entry(best_name, is_micropost):
    entry = TestEntry()
    entry.best_name = best_name
    entry.is_micropost = is_micropost
    entry.url = test_url
    return entry


@pytest.mark.parametrize("title, limit, add_url, expected", [
    ('A test entry', None, False, 'A test entry'),
    ('A test entry', None, 'auto', 'A test entry ' + test_url),
    ('A test entry', None, True, 'A test entry ' + test_url),

    ('A test entry', 80, False, 'A test entry'),
    ('A test entry', 80, 'auto', 'A test entry ' + test_url),
    ('A test entry', 80, True, 'A test entry ' + test_url),

    ('A test entry that is very very long because its title has many many '
     'words in it for no good reason', 80, False,
     'A test entry that is very very long because its title has many many '
     'words in...'),
    ('A test entry that is very very long because its title has many many '
     'words in it for no good reason', 80, 'auto',
     'A test entry that is very very long because its... ' + test_url),
    ('A test entry that is very very long because its title has many many '
     'words in it for no good reason', 80, True,
     'A test entry that is very very long because its... ' + test_url)
])
def test_format_lonform_entry(title, limit, add_url, expected):
    entry = _make_test_entry(title, False)
    actual = format_entry(entry, limit, add_url)
    assert actual == expected


@pytest.mark.parametrize("text, limit, add_url, expected", [
    ('A test entry', None, False, 'A test entry'),
    ('A test entry', None, 'auto', 'A test entry'),
    ('A test entry', None, True, 'A test entry ' + test_url),

    ('A test entry', 80, False, 'A test entry'),
    ('A test entry', 80, 'auto', 'A test entry'),
    ('A test entry', 80, True, 'A test entry ' + test_url),

    ('A test entry that is very very long because its title has many many '
     'words in it for no good reason', 80, False,
     'A test entry that is very very long because its title has many many '
     'words in...'),
    ('A test entry that is very very long because its title has many many '
     'words in it for no good reason', 80, 'auto',
     'A test entry that is very very long because its... ' + test_url),
    ('A test entry that is very very long because its title has many many '
     'words in it for no good reason', 80, True,
     'A test entry that is very very long because its... ' + test_url)
])
def test_format_micropost_entry(text, limit, add_url, expected):
    entry = _make_test_entry(text, True)
    actual = format_entry(entry, limit, add_url)
    assert actual == expected



import re
import textwrap


def format_entry(entry, limit=None, add_url='auto'):
    url = entry.url
    name = entry.best_name

    do_add_url = ((add_url is True) or
                  (add_url == 'auto' and not entry.is_micropost))
    if limit:
        if do_add_url and url:
            limit -= 1 + len(url)

        shortened = len(name) > limit
        if shortened:
            # If we have to shorten the text, but we haven't taken the
            # URL into account yet, let's see if we have to include now!
            # (this happens when we only want to include it when the text
            #  is shortened)
            if not do_add_url and add_url == 'auto' and url:
                do_add_url = True
                limit -= 1 + len(url)

        if limit <= 0:
            raise Exception("Can't shorten post name.")

        name = textwrap.shorten(name, width=limit, placeholder="...")

    if do_add_url and url:
        name += ' ' + url
    return name


re_sentence_end = re.compile(r'[\w\]\)\"\'\.]\.\s|[\?\!]\s')


def shorten_text(txt, limit):
    if len(txt) <= limit:
        return (txt, False)

    m = re_sentence_end.search(txt)
    if m and m.end <= (limit + 1):
        return (txt[:m.end - 1], True)

    shorter = textwrap.shorten(
        txt, width=limit, placeholder="...")
    return (shorter, True)




feed1 = """
<html><body>
    <article class="h-entry">
      <h1 class="p-name">A new article</h1>
      <div class="e-content">
        <p>This is the text of the article.</p>
        <p>It has 2 paragraphs.</p>
      </div>
      <a class="u-url" href="https://example.org/a-new-article">permalink</a>
    </article>
</body></html>"""


def test_populate(cli):
    cli.appendSiloConfig('test', 'print', items='name')
    feed = cli.createTempFeed(feed1)
    ctx, _ = cli.run('populate', feed, '-s', 'test')
    assert ctx.cache.wasPosted('test', 'https://example.org/a-new-article')


feed2 = """
<html><body>
    <article class="h-entry">
      <h1 class="p-name">First article</h1>
      <div><time class="dt-published" datetime="2018-01-07T09:30:00-00:00"></time></div>
      <a class="u-url" href="https://example.org/first-article">permalink</a>
    </article>
    <article class="h-entry">
      <h1 class="p-name">Second article</h1>
      <div><time class="dt-published" datetime="2018-01-08T09:30:00-00:00"></time></div>
      <a class="u-url" href="https://example.org/second-article">permalink</a>
    </article>
    <article class="h-entry">
      <h1 class="p-name">Third article</h1>
      <div><time class="dt-published" datetime="2018-01-09T09:30:00-00:00"></time></div>
      <a class="u-url" href="https://example.org/third-article">permalink</a>
    </article>
</body></html>"""  # NOQA


def test_populate_until(cli):
    cli.appendSiloConfig('test', 'print', items='name')
    feed = cli.createTempFeed(feed2)
    ctx, _ = cli.run('populate', feed, '-s', 'test', '--until', '2018-01-08')
    assert ctx.cache.wasPosted('test', 'https://example.org/first-article')
    assert ctx.cache.wasPosted('test', 'https://example.org/second-article')
    assert not ctx.cache.wasPosted('test', 'https://example.org/third-article')



import datetime
import textwrap
from .base import Silo


class PrintSilo(Silo):
    SILO_TYPE = 'print'

    def authenticate(self, ctx):
        pass

    def postEntry(self, entry, ctx):
        import pprint

        tokens = {}
        shorten = (self.getConfigItem('shorten', '').lower() in
                   ['true', 'yes', 'on', '1'])
        names = self.getConfigItem('items', 'best_name,category,published')
        names = names.split(',')
        for n in names:
            if n == 'type':
                tokens['type'] = entry.entry_type

            elif n == 'best_name':
                tokens['best_name'] = entry.best_name

            else:
                v = entry.get(n)
                if shorten:
                    v = textwrap.shorten(v, width=400, placeholder='...')
                if isinstance(v, datetime.datetime):
                    v = v.strftime('%c')
                tokens[n] = v

        pprint.pprint(tokens)



import os.path
from setuptools import setup, find_packages


def read(fname):
    with open(os.path.join(os.path.dirname(__file__), fname)) as fp:
        return fp.read()


long_description = read('README.rst')

install_requires = [
    'coloredlogs>=10.0',
    'Mastodon.py>=1.3.0',
    'mf2py>=1.1.0',
    'mf2util>=0.5.0',
    'python-dateutil>=2.7.0',
    'python-twitter>=3.4.0'
]

tests_require = [
    'pytest>=3.6.2'
]

setup_requires = [
    'setuptools-scm',
    'pytest-runner'
]


setup(
    name='silorider',
    use_scm_version={'write_to': 'silorider/version.py'},
    description=("Scans a website's microformats and cross-posts content "
                 "to 'silo' services."),
    long_description=long_description,
    author='Ludovic Chabant',
    author_email='ludovic@chabant.com',
    license="Apache License 2.0",
    url='https://bolt80.com/silorider',
    packages=find_packages(),
    setup_requires=setup_requires,
    tests_require=tests_require,
    install_requires=install_requires,
    entry_points={'console_scripts': [
        'silorider = silorider.main:main'
    ]}
)




def mock_urllib(m):
    import urllib.request
    m.setattr(urllib.request, 'urlretrieve', _patched_urlretrieve)
    m.setattr(urllib.request, 'urlcleanup', _patched_urlcleanup)
    return m


def _patched_urlretrieve(url):
    return ('/retrieved/' + url.lstrip('/'), None)


def _patched_urlcleanup():
    pass



import logging
from .utils import get_named_silos
from ..silos.base import SiloPostingContext
from ..parse import parse_url


logger = logging.getLogger(__name__)


def process_url(url, ctx):
    p = Processor(ctx, url)
    p.process()


class Processor:
    def __init__(self, ctx, url):
        self.ctx = ctx
        self.url = url
        self._silos = get_named_silos(ctx.silos, ctx.args.silo)

    @property
    def config(self):
        return self.ctx.config

    @property
    def silos(self):
        return self._silos

    def process(self):
        self.preProcess()

        feed = parse_url(self.url)
        for entry in feed.entries:
            self.processEntry(entry)

        self.postProcess()

    def preProcess(self):
        for silo in self.silos:
            silo.onPostStart()

    def postProcess(self):
        for silo in self.silos:
            silo.onPostEnd()

    def processEntry(self, entry):
        if self.isEntryFiltered(entry):
            logger.debug("Entry is filtered out: %s" % entry.best_name)
            return

        entry_url = entry.get('url')
        if not entry_url:
            logger.warning("Found entry without a URL.")
            return

        postctx = SiloPostingContext(self.ctx)
        no_cache = self.ctx.args.no_cache
        logger.debug("Processing entry: %s" % entry.best_name)
        for silo in self.silos:
            if no_cache or not self.ctx.cache.wasPosted(silo.name, entry_url):
                if not self.ctx.args.dry_run:
                    silo.postEntry(entry, postctx)
                    self.ctx.cache.addPost(silo.name, entry_url)
                else:
                    logger.info("Would post entry on %s: %s" %
                                (silo.name, entry.best_name))
            else:
                logger.debug("Skipping already posted entry on %s: %s" %
                             (silo.name, entry.best_name))

    def isEntryFiltered(self, entry):
        if not self.config.has_section('filter'):
            return False

        items = self.config.items('filter')

        for name, value in items:
            if name.startswith('include_'):
                propname = name[8:]
                propvalue = entry.get(propname)
                for inc_val in value.split(','):
                    if inc_val in propvalue:
                        break
                else:
                    return True

            elif name.startswith('exclude_'):
                propname = name[8:]
                propvalue = entry.get(propname)
                for excl_val in value.split(','):
                    if excl_val in propvalue:
                        return True

        return False
